/* @(#)scsi-next.c	1.7 98/11/02 Copyright 1997 J. Schilling */
#ifndef lint
static	char __sccsid[] =
	"@(#)scsi-next.c	1.7 98/11/02 Copyright 1997 J. Schilling";
#endif
/*
 *	Interface for the NeXT Step generic SCSI implementation.
 *
 *	This is a hack, that tries to emulate the functionality
 *	of the scg driver.
 *
 *	Copyright (c) 1997 J. Schilling
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <bsd/dev/scsireg.h>

#define	MAX_SCG		4	/* Max # of SCSI controllers */
#define	MAX_TGT		8
#define	MAX_LUN		8

LOCAL	int	scgfile;
LOCAL	int	max_scsibus	= -1;
LOCAL	int	cur_scsibus	= -1;
LOCAL	int	cur_target	= -1;
LOCAL	int	cur_lun		= -1;

/*#define	MAX_DMA_NEXT	(32*1024)*/
#define	MAX_DMA_NEXT	(64*1024)	/* Check if this is not too big */


LOCAL	BOOL	scsi_setup	__PR((SCSI *scgp, int f, int busno, int tgt, int tlun,
								BOOL ex));

EXPORT
int scsi_open(scgp, device, busno, tgt, tlun)
	SCSI	*scgp;
	char	*device;
	int	busno;
	int	tgt;
	int	tlun;
{
	register int	f;
	register int	i;
	char		devname[64];

	if ((device != NULL && *device != '\0') || (busno == -2 && tgt == -2))
		return (-1);

	if (busno >= MAX_SCG || tgt >= MAX_TGT || tlun >= MAX_LUN)
		return (-1);

	for (i=0; i < 4; i++) {
		sprintf(devname, "/dev/sg%d", i);
		f = open(devname, O_RDWR);
		if (scgp->debug)
			errmsg("open(devname: '%s') : %d\n", devname, f);
		if (f < 0)
			continue;
		scgfile = f;
		break;

	}
	if (f >= 0) {
		if (max_scsibus < 0) {
			for (i=0; i < 4; i++) {
				if (!scsi_havebus(i))
					break;
			}
			max_scsibus = i;
		}
		if (scgp->debug)
			printf("maxbus: %d\n", max_scsibus);
		if (max_scsibus <= 0) {
			max_scsibus = 1;
			cur_scsibus = 0;
		}

       	        ioctl(f, SGIOCENAS);
		if (busno > 0 && tgt > 0 && tlun > 0)
			scsi_setup(scgp, f, busno, tgt, tlun, TRUE);
		return(1);
	}
	return (0);
}

LOCAL BOOL
scsi_setup(scgp, f, busno, tgt, tlun, ex)
	SCSI	*scgp;
	int	f;
	int	busno;
	int	tgt;
	int	tlun;
	BOOL	ex;
{
	scsi_adr_t sadr;

	sadr.sa_target = tgt;
	sadr.sa_lun = tlun;

/*printf("curbus%d %d\n", cur_scsibus, busno);*/
/*	if ((cur_scsibus < 0 || cur_scsibus != busno))*/
/*		printf("set\n");*/
	if ((cur_scsibus < 0 || cur_scsibus != busno) &&
				ioctl(f, SGIOCCNTR, &busno) < 0) {
		if (ex)
/*			comerr("Cannot set SCSI bus\n");*/
			errmsg("Cannot set SCSI bus\n");
		return (FALSE);
	}
	cur_scsibus	= busno;

	if (ioctl(f, SGIOCSTL, &sadr) < 0) {
		if (ex)
			comerr("Cannot set SCSI address\n");
		return (FALSE);
	}
	cur_scsibus	= busno;
	cur_target	= tgt;
	cur_lun		= tlun;
	return (TRUE);
}

LOCAL long
scsi_maxdma(scgp)
	SCSI	*scgp;
{
	long maxdma = MAX_DMA_NEXT;
#ifdef	SGIOCMAXDMA
	int  m;

	if (ioctl(scgfile, SGIOCMAXDMA, &m) >= 0)
		maxdma = m;
		if (scgp->debug)
			printf("maxdma: %d\n", maxdma);
#endif
	return (maxdma);
}
#ifdef	XXX
#define SGIOCENAS	_IO(  's', 2)		 	/* enable autosense */
#define SGIOCDAS	_IO(  's', 3)			/* disable autosense */
#define SGIOCRST	_IO(  's', 4)			/* reset SCSI bus */
#define SGIOCCNTR       _IOW( 's', 6, int)              /* select controller */
#define SGIOCGAS	_IOR( 's', 7, int)		/* get autosense */
#define SGIOCMAXDMA	_IOR( 's', 8, int)		/* max DMA size */
#define SGIOCNUMTARGS	_IOR( 's', 9, int)		/* # of targets/bus */
#endif

EXPORT void *
scsi_getbuf(scgp, amt)
	SCSI	*scgp;
	long	amt;
{
	void	*ret;

	if (scg_maxdma == 0)
		scg_maxdma = scsi_maxdma(scgp);

	if (amt <= 0 || amt > scg_maxdma)
		return ((void *)0);
	if (scgp->debug)
		printf("scsi_getbuf: %ld bytes\n", amt);
	ret = valloc((size_t)(amt));
	return (ret);
}

EXPORT
BOOL scsi_havebus(scgp, busno)
	SCSI	*scgp;
	int	busno;
{
	if (busno < 0 || busno >= MAX_SCG)
		return (FALSE);
	if (max_scsibus > 0 && busno >= max_scsibus)
		return (FALSE);

	return (scsi_setup(scgp, scgfile, busno, 0, 0, FALSE));
}

EXPORT
int scsi_fileno(scgp, busno, tgt, tlun)
	SCSI	*scgp;
	int	busno;
	int	tgt;
	int	tlun;
{
	if (busno < 0 || busno >= MAX_SCG ||
	    tgt < 0 || tgt >= MAX_TGT ||
	    tlun < 0 || tlun >= MAX_LUN)
		return (-1);
	if (max_scsibus > 0 && busno >= max_scsibus)
		return (-1);

	if ((busno != cur_scsibus) || (tgt != cur_target) || (tlun != cur_lun)) {
		if (!scsi_setup(scgfile, busno, tgt, tlun, FALSE))
			return (-1);
	}
	return (scgfile);
}

EXPORT
int scsi_isatapi(scgp)
	SCSI	*scgp;

{
	return (FALSE);
}

EXPORT
int scsireset(scgp)
	SCSI	*scgp;
{
	int	f = scsi_fileno(scgp, scgp->scsibus, scgp->target, scgp->lun);

	return (ioctl(f, SGIOCRST, 0));
}

LOCAL int
scsi_send(scgp, f, sp)
	SCSI		*scgp;
	int		f;
	struct scg_cmd	*sp;
{
	struct scsi_req	req;
	register long	*lp1;
	register long	*lp2;
	int		ret = 0;

	if (f < 0 || (sp->cdb_len > sizeof(req.sr_cdb))) {
		sp->error = SCG_FATAL;
		sp->ux_errno = EIO;
		return (0);
	}
	fillbytes(&req, sizeof(req), '\0');
	movebytes(sp->cdb.cmd_cdb, &req.sr_cdb, sp->cdb_len);
	if (sp->size) {
		req.sr_dma_dir = SR_DMA_WR;
		if (sp->flags & SCG_RECV_DATA)
			req.sr_dma_dir = SR_DMA_RD;
	}
	req.sr_addr = sp->addr;
	req.sr_dma_max = sp->size;
	req.sr_ioto = sp->timeout;
	if (ioctl(f, SGIOCREQ, (void *)&req) < 0) {
		ret  = -1;
		sp->ux_errno = geterrno();
		if (sp->ux_errno != ENOTTY)
			ret = 0;
	} else {
		sp->ux_errno = 0;
	}
	sp->u_scb.cmd_scb[0] = req.sr_scsi_status;
	sp->sense_count = sizeof(esense_reply_t);
	if (sp->sense_count > sp->sense_len)
		sp->sense_count = sp->sense_len;
	if (sp->sense_count > SCG_MAX_SENSE)
		sp->sense_count = SCG_MAX_SENSE;
	if (sp->sense_count < 0)
		sp->sense_count = 0;
	movebytes(&req.sr_esense, sp->u_sense.cmd_sense, sp->sense_count);
	sp->resid = sp->size - req.sr_dma_xfr;

	switch (req.sr_io_status) {

	case SR_IOST_GOOD:	sp->error = SCG_NO_ERROR;	break;

	case SR_IOST_CHKSNV:	sp->sense_count = 0;
	case SR_IOST_CHKSV:	sp->error = SCG_RETRYABLE;
				break;
	
	case SR_IOST_SELTO:
	case SR_IOST_DMAOR:
				sp->error = SCG_FATAL;		break;

	case SR_IOST_IOTO:	sp->error = SCG_TIMEOUT;	break;

	case SR_IOST_PERM:
	case SR_IOST_NOPEN:
				sp->error = SCG_FATAL;
				ret = (-1);
				break;

	default:		sp->error = SCG_RETRYABLE;	break;

	}
	return (ret);
}
